//_GRAPHICS PROGRAMMING COLUMN_
//by Michael Abrash
//may91a.c

//[LISTING ONE]

/* Animates a line rotating about its center. 180 frames are constructed 
offscreen, one for each 1 degree rotation, and then copied to the screen to 
produce animation. Edsun CEG antialiasing or emulated CEG antialiasing may 
optionally be performed. To compile for Edsun CEG antialiasing, define 
USE_CEG on the compiler command line (/DUSE_CEG for MSC, -DUSE_CEG for Turbo C)
and link this code with Listing 3. To compile for emulated CEG antialiasing, 
define EMULATE_CEG on the compiler command line and link this code with 
Listing 4. All C code tested with Microsoft C 5.0. Requires a large data model
(the compact model was used for testing). */

#include <stdio.h>
#include <stdlib.h>
#include <conio.h>
#include <math.h>
#include <dos.h>
#ifdef __TURBOC__
#include <mem.h>
#else
#include <memory.h>
#endif

/* Size of frames drawn to system memory, and size of the circle
   formed by the rotation of the line segment around its center */
#define FRAME_WIDTH     49
#define FRAME_HEIGHT    49
#define FRAME_CENTER_X  (FRAME_WIDTH/2)
#define FRAME_CENTER_Y  (FRAME_HEIGHT/2)
#define X_RADIUS        (FRAME_WIDTH/2-1)
#define Y_RADIUS        (FRAME_HEIGHT/2-1)
#define SCREEN_WIDTH    320
#define SCREEN_HEIGHT   200
#define SCREEN_SEGMENT  0xA000
#define X_OFFSET_ADJUST ((SCREEN_WIDTH/2) - FRAME_CENTER_X)
#define Y_OFFSET_ADJUST ((SCREEN_HEIGHT/2) - FRAME_CENTER_Y)
#define PI              3.141592

int main(void);
extern void DrawLine(unsigned char *,int,int,int,int,int,int,int);
extern int SetCEGMode(int);
extern void SetAAPalette(void);

int main() {
   unsigned char *FramePtr[180], *ScreenPtr, *TempPtr;
   int angle, X1, Y1, X2, Y2, temp, i;
   union REGS regset;

   /* Generate 180 frames, one for each 1 degree rotation of the line
      segment about its center; store the results in system memory */
   printf("Precalculating frames. Please wait...");
   /* First, allocate space for the frames */
   for (angle = 0; angle < 180; angle++) {
      if ((FramePtr[angle] = (unsigned char *)
            malloc(FRAME_WIDTH * FRAME_HEIGHT)) == NULL) {
         printf("Out of memory\n");
         return(1);
      }
      /* Clear the frame to black */
      memset(FramePtr[angle], 0, FRAME_WIDTH * FRAME_HEIGHT);
   }

   /* Generate the frames, one for each 1 degree rotation of the line
      segment about its center */
   for (angle = 0; angle < 180; angle++) {
      /* Calculate upper end of line as a counterclockwise rotation from right
         end of a horizontal line and lower end of line as a counterclockwise 
         rotation from left end of a horizontal line */
      temp = cos((double)angle * PI / 180.0) * X_RADIUS + 0.5;
      X1 = FRAME_CENTER_X + temp;
      X2 = FRAME_CENTER_X - temp;
      temp = sin((double)angle * PI / 180.0) * Y_RADIUS + 0.5;
      Y1 = FRAME_CENTER_Y - temp;
      Y2 = FRAME_CENTER_Y + temp;
      /* Draw the line in white */
      DrawLine(FramePtr[angle], FRAME_WIDTH, FRAME_HEIGHT, X1, Y1,
            X2, Y2, 15);
   }

   /* Set to the standard 256-color VGA mode, mode 0x13, 320x200 */
   regset.x.ax = 0x0013; int86(0x10, &regset, &regset);

#ifdef USE_CEG
   /* Enable Advanced-8 CEG mode */
   if (!SetCEGMode(13)) {
   /* Restore text mode and we're done */
      regset.x.ax = 0x0003;
      int86(0x10, &regset, &regset);
      fprintf(stderr, "CEG/DAC not installed\n");
      return(1);     /* no CEG/DAC installed */
   }
#endif

#ifdef EMULATE_CEG
   /* Set the palette up for antialiasing */
   SetAAPalette();
#endif

   /* Draw the frames, at a rate of 1 frame per screen refresh
      interval, until a key is pressed */
   for (angle = 0;;) {
    do {
      /* Point to the destination area of display memory */
#ifdef __TURBOC__
      ScreenPtr = MK_FP(SCREEN_SEGMENT, (Y_OFFSET_ADJUST*SCREEN_WIDTH)
            + X_OFFSET_ADJUST);
#else
      FP_SEG(ScreenPtr) = SCREEN_SEGMENT;
      FP_OFF(ScreenPtr) =
            (Y_OFFSET_ADJUST * SCREEN_WIDTH) + X_OFFSET_ADJUST;
#endif
      /* Wait for the start of the vertical non-display portion of the
         frame */
      while (inp(0x3DA) & 0x08) ;
      while (!(inp(0x3DA) & 0x08)) ;
      /* Copy over the current frame, a scan line at a time */
      for (i = 0, TempPtr = FramePtr[angle]; i < FRAME_HEIGHT;
            i++, ScreenPtr += SCREEN_WIDTH, TempPtr += FRAME_WIDTH) {
         memcpy(ScreenPtr, TempPtr, FRAME_WIDTH);
      }
      angle = (angle + 1) % 180; /* wrap back every 180 frames */
    } while (!kbhit());
    if (getch() == 0x1B)   /* pause if a key was pressed */
      break;               /* exit if the key was Esc */
    getch();               /* wait for another key and resume */
   }

   /* Return the CEG/DAC to standard VGA operation by writing to
      palette location 223, restore text mode and we're done */
#ifdef USE_CEG
   while (inp(0x3DA) & 0x08) ;      /* wait for the start of */
   while (!(inp(0x3DA) & 0x08)) ;   /* vertical non-display */
   outp(0x3C8, 223); outp(0x3C9, 0); outp(0x3C9, 0); outp(0x3C9, 0);
#endif
   regset.x.ax = 0x0003; int86(0x10, &regset, &regset);
   return(0);
}

