//[LISTING THREE]

/* Draws an Advanced-8 CEG-antialiased line from (X1,Y1) to (X2,Y2) into buffer
pointed to by BufferPtr, of width BufferWidth and height BufferHeight. */

#include <dos.h>
#include <math.h>
#define SWAP(a,b) {temp = a; a = b; b = temp;}

void DrawLine(unsigned char *BufferPtr, int BufferWidth,
      int BufferHeight, int X1, int Y1, int X2, int Y2, int color)
{
   int X, Y, DeltaX, DeltaY, temp, WeightingIndex, i, MixLength;
   double Slope, InverseSlope, XFloat, YFloat;

   /* Calculate X and Y lengths of the line */
   DeltaX = X2 - X1;
   DeltaY = Y2 - Y1;

   /* Determine the major axis */
   if (abs(DeltaY) > abs(DeltaX)) {
      /* Y is the major axis */
      if (DeltaY < 0) {    /* make sure DeltaY is positive */
         SWAP(X1, X2);
         SWAP(Y1, Y2);
         DeltaX = -DeltaX;
         DeltaY = -DeltaY;
      }
      InverseSlope = (double)DeltaX / (double)DeltaY;
      /* Scan out line, stepping along the Y axis 1 pixel at a time and 
         calculating corresponding pair of X coordinates, 1 on each side of 
         ideal line, with the mix between background color and line color at 
         each of the two X coordinates proportional to proximity of line to 
         that pixel, and with line color intensities of the two X coordinates 
         summing to 100% */
      for (Y = Y1; Y <= Y2; Y++) {
         /* Exact X coordinate at this Y coordinate */
         XFloat = (double)X1 + ((double)(Y - Y1) * InverseSlope);
         /* Nearest X coordinate on or to the left of the line */
         X = (int)floor(XFloat);
         /* Draw the color to the left pixel */
         *(BufferPtr + BufferWidth * Y + X) = color;
         /* Draw the weighting index for the desired color mix for the
            left pixel to the right pixel; the CEG/DAC uses this to
            mix the color we just wrote with the background color and
            draw that for the left pixel, then uses the complementary
            mix to draw the right pixel (this pixel). Confusing, but
            that's the way the CEG/DAC works! */
         *(BufferPtr + BufferWidth * Y + X + 1) =
               (int)((XFloat - X) * 32.0) + 192;
      }
   } else {
      /* X is the major axis */
      if (DeltaX < 0) {    /* make sure DeltaX is positive */
         SWAP(X1, X2);
         SWAP(Y1, Y2);
         DeltaX = -DeltaX;
         DeltaY = -DeltaY;
      }
      Slope = (double)DeltaY / (double)DeltaX;
      /* Scan out the line, stepping along the X axis one pixel at a
         time and calculating the corresponding pair of Y coordinates,
         one on each side of the ideal line, with the mix between the
         background color and the line color at each of the two Y
         coordinates proportional to the proximity of the line to that
         pixel, and with the line color intensities of the two Y
         coordinates summing to 100% */
      for (X = X1; X <= X2; X++) {
         /* Exact Y coordinate at this X coordinate */
         YFloat = (double)Y1 + ((double)(X - X1) * Slope);
         /* Nearest Y coordinate on or above the line */
         Y = (int)floor(YFloat);
         /* Calculate the weighting index for the percentage of this
            pixel that's on the top scan line of the pair this pixel
            is split between */
         WeightingIndex = (int)((YFloat - Y) * 32.0);
         /* Set the weighting for the top pixel */
         *(BufferPtr + BufferWidth * Y + X) =
               WeightingIndex + 192;
         /* Set the weighting for the bottom pixel, with the top and
            bottom weightings summing to 100% */
         *(BufferPtr + BufferWidth * (Y + 1) + X) =
               31 - WeightingIndex + 192;
      }

      /* Finally, post-process the buffer to put leading mix color
         bytes on mix sequences (so the weighting indexes have
         something to mix), and to turn one-pixel-wide sequences
         (artifacts of the above drawing approach, which will not
         display properly) into two-pixel-wide sequences by appending
         an additional mixed pixel 100% weighted to the background color */
      for (i = 0, MixLength = 0; i < (BufferWidth*BufferHeight); i++) {
         if (*(BufferPtr + i) != 0) {
            /* Part of a mix sequence; increment sequence length */
            MixLength++;
         } else {
            if (MixLength > 0) {
               /* Mix sequence just ended; set the line color to start
                  the mix sequence */
                  *(BufferPtr + i - MixLength - 1) = color;
               if (MixLength == 1)
                  /* This is a 1-long mix sequence; pad it to a 2-long
                     sequence with an all-background mixed pixel, so
                     the mix sequence will display properly, rather
                     than as a special 2-wide color/mix line case */
                  *(BufferPtr + i) = 0xDF;
               MixLength = 0;
            }
         }
      }
   }
}

/* Sets the desired CEG/DAC mode, enabling CEG graphics. Returns 1 for
   success, 0 if no CEG/DAC is installed. */
int SetCEGMode(int mode) {
   /* Wait for the start of the vertical non-display portion of the
      frame */
   while (inp(0x3DA) & 0x08) ;
   while (!(inp(0x3DA) & 0x08)) ;

   outp(0x3C7, 222);    /* write the CEG enable sequence */
   outp(0x3C9, 'C'); outp(0x3C9, 'E'); outp(0x3C9, 'G');
   outp(0x3C7, 222);
   outp(0x3C9, 'E'); outp(0x3C9, 'D'); outp(0x3C9, 'S');
   outp(0x3C7, 222);
   outp(0x3C9, 'U'); outp(0x3C9, 'N');
   outp(0x3C9, mode);   /* write the CEG mode */

   /* CEG mode should be enabled. Make sure this is a CEG/DAC */
   outp(0x3C6, 0xFF);   /* enable all DAC mask bits */
   if ((inp(0x3C6) & 0x70) == 0x70)
      return(0);  /* no version # bit is 0; this is not a CEG/DAC */
   else
      return(1);  /* this is a CEG/DAC, and it's ready to go */
}
