//[LISTING FOUR]

/* Draws an antialiased line from (X1,Y1) to (X2,Y2) into the buffer pointed 
to by BufferPtr, of width BufferWidth and height BufferHeight. White on black 
is the only supported color combination. */

#include <dos.h>
#include <math.h>
#define SWAP(a,b) {temp = a; a = b; b = temp;}

void DrawLine(unsigned char *BufferPtr, int BufferWidth,
      int BufferHeight, int X1, int Y1, int X2, int Y2, int color)
{
   int X, Y, DeltaX, DeltaY, temp, WeightingIndex, i, MixLength;
   double Slope, InverseSlope, XFloat, YFloat;

   /* Calculate X and Y lengths of the line */
   DeltaX = X2 - X1;
   DeltaY = Y2 - Y1;

   /* Determine the major axis */
   if (abs(DeltaY) > abs(DeltaX)) {
      /* Y is the major axis */
      if (DeltaY < 0) {    /* make sure DeltaY is positive */
         SWAP(X1, X2);
         SWAP(Y1, Y2);
         DeltaX = -DeltaX;
         DeltaY = -DeltaY;
      }
      InverseSlope = (double)DeltaX / (double)DeltaY;
      /* Scan out line, stepping along the Y axis one pixel at a time and 
         calculating corresponding pair of X coordinates, 1 on each side of 
         ideal line, with the mix between background color and line color at 
         each of the two X coordinates proportional to proximity of line to 
         that pixel, and with line color intensities of two X coordinates 
         summing to 100% */
      for (Y = Y1; Y <= Y2; Y++) {
         /* Exact X coordinate at this Y coordinate */
         XFloat = (double)X1 + ((double)(Y - Y1) * InverseSlope);
         /* Nearest X coordinate on or to the left of the line */
         X = (int)floor(XFloat);
         /* Calculate the weighting index for the percentage of this
            pixel that's in the left column of the pair this pixel is
            split between */
         WeightingIndex = (int)((XFloat - X) * 32.0);
         /* Draw the left pixel with the desired color weighting */
         *(BufferPtr + BufferWidth * Y + X) = WeightingIndex + 16;
         /* Draw the right pixel with the complement of the left pixel
            color weighting, so that the line color intensities sum to 100% */
         *(BufferPtr + BufferWidth * Y + X + 1) =
               31 - WeightingIndex + 16;
      }
   } else {
      /* X is the major axis */
      if (DeltaX < 0) {    /* make sure DeltaX is positive */
         SWAP(X1, X2);
         SWAP(Y1, Y2);
         DeltaX = -DeltaX;
         DeltaY = -DeltaY;
      }
      /* Scan out the line, stepping along the X axis one pixel at a
         time and calculating the corresponding pair of Y coordinates,
         one on each side of the ideal line, with the mix between the
         background color and the line color at each of the two Y
         coordinates proportional to the proximity of the line to that
         pixel, and with the line color intensities of the two Y
         coordinates summing to 100% */
      Slope = (double)DeltaY / (double)DeltaX;
      for (X = X1; X <= X2; X++) {
         /* Exact Y coordinate at this X coordinate */
         YFloat = (double)Y1 + ((double)(X - X1) * Slope);
         /* Nearest Y coordinate on or above the line */
         Y = (int)floor(YFloat);
         /* Calculate the weighting index for the percentage of this
            pixel that's on the top scan line of the pair this pixel
            is split between */
         WeightingIndex = (int)((YFloat - Y) * 32.0);
         /* Draw the top pixel with the desired color weighting */
         *(BufferPtr + BufferWidth * Y + X) =
               WeightingIndex + 16;
         /* Draw the bottom pixel with the complement of the top pixel
            color weighting, so that the line color intensities sum to
            100% */
         *(BufferPtr + BufferWidth * (Y + 1) + X) =
               31 - WeightingIndex + 16;
      }
   }
}

/* Sets palette entries 16-47 for antialiasing, stepping from solid white to 
solid black (increments of 1/31st.) Steps are corrected for a gamma of 2.3. */
void SetAAPalette() {
   union REGS regset;
   struct SREGS sregs;
   static unsigned char AASettings[32*3] = {
      63,63,63, 62,62,62, 61,61,61, 60,60,60, 59,59,59, 58,58,58,
      57,57,57, 56,56,56, 55,55,55, 54,54,54, 53,53,53, 52,52,52,
      51,51,51, 50,50,50, 49,49,49, 47,47,47, 46,46,46, 45,45,45,
      43,43,43, 42,42,42, 40,40,40, 39,39,39, 37,37,37, 35,35,35,
      33,33,33, 31,31,31, 28,28,28, 26,26,26, 23,23,23, 19,19,19,
      14,14,14,  0, 0, 0
   };

   regset.x.ax = 0x1012;
   regset.x.bx = 0x0010;
   regset.x.cx = 0x0020;
   regset.x.dx = (unsigned int) AASettings;
   segread(&sregs);
   sregs.es = sregs.ds;       /* point ES:DX to AASettings */
   int86x(0x10, &regset, &regset, &sregs);
}

